import logging
import os
from datetime import datetime
from enum import Enum
from typing import Optional

from pydantic import BaseModel, Field, HttpUrl, SecretStr, ValidationError, validator

from ..DataGoKr import DataGoKr


# logging
logger = logging.getLogger(__file__)

# debug only
AIRKOREA_API_KEY = os.getenv("AIRKOREA_API_KEY")

################################################################################
# Types
################################################################################
class ReturnType(str, Enum):
    # Only JSON Available yet
    json = "JSON"


# (Type)
class DataTerm(str, Enum):
    DAILY = "DAILY"
    MONTH = "MONTH"
    _3MONTH = "3MONTH"


################################################################################
# [Abstract] Abstract for VilageFcst
################################################################################
class ArpltnInforInqireSvc(DataGoKr):
    __version__ = "1.7.2"

    baseUrl: HttpUrl = "http://apis.data.go.kr/B552584/ArpltnInforInqireSvc"
    returnType: ReturnType = "JSON"
    serviceKey: str = AIRKOREA_API_KEY


################################################################################
# [API] 측정소별 실시간 측정정보 조회
#
# 버전(ver) 항목설명
# - 버전을 포함하지 않고 호출할 경우 : PM2.5 데이터가 포함되지 않은 원래 오퍼레이션 결과 표출.
# - 버전 1.0을 호출할 경우 : PM2.5 데이터가 포함된 결과 표출.
# - 버전 1.1을 호출할 경우 : PM10, PM2.5 24시간 예측이동 평균데이터가 포함된 결과 표출.
# - 버전 1.2을 호출할 경우 : 측정망 정보 데이터가 포함된 결과 표출.
# - 버전 1.3을 호출할 경우 : PM10, PM2.5 1시간 등급 자료가 포함된 결과 표출
#
# Grade 값의 의미
#   1: 좋음, 2: 보통, 3: 나쁨, 4: 매우나쁨
################################################################################
# Output Model
class MsrstnAcctoRltmMesureDnstyModel(BaseModel):
    dataTime: str
    mangName: str
    so2Value: Optional[float]
    coValue: Optional[float]
    o3Value: Optional[float]
    no2Value: Optional[float]
    pm10Value: Optional[int]
    pm10Value24: Optional[int]
    pm25Value: Optional[int]
    pm25Value24: Optional[int]
    khaiValue: Optional[int]
    khaiGrade: Optional[int]
    so2Grade: Optional[int]
    coGrade: Optional[int]
    o3Grade: Optional[int]
    no2Grade: Optional[int]
    pm10Grade: Optional[int]
    pm25Grade: Optional[int]
    so2Flag: Optional[str]
    coFlag: Optional[str]
    o3Flag: Optional[str]
    no2Flag: Optional[str]
    pm10Flag: Optional[str]
    pm25Flag: Optional[str]

    @validator("*", pre=True)
    def _hyphen_to_none(cls, v):
        """
        (NOTE)
        Airkorea API는 자료이상 시 null이 아닌 string "-"을 반환하여 schema vaidate error 발생.
        pre-validator에서 "-"을 None으로 변환하여 해결.
        """
        if v in ["-"]:
            return None
        return v


# API
class MsrstnAcctoRltmMesureDnsty(ArpltnInforInqireSvc):
    __RecordModel__ = MsrstnAcctoRltmMesureDnstyModel
    __index_names__ = None
    __key_name__ = None
    __value_name__ = None

    route: str = "getMsrstnAcctoRltmMesureDnsty"
    stationName: str = "종로구"
    dataTerm: str = "DAILY"
    ver: str = "1.3"


################################################################################
# [API] 통합대기환경지수 나쁨 이상 측정소 목록조회
################################################################################
# Output Model
class UnityAirEnvrnIdexSnstiveAboveMsrstnListModel(BaseModel):
    stationName: str
    addr: str


# API
class UnityAirEnvrnIdexSnstiveAboveMsrstnList(ArpltnInforInqireSvc):
    __RecordModel__ = UnityAirEnvrnIdexSnstiveAboveMsrstnListModel
    __index_names__ = None
    __key_name__ = None
    __value_name__ = None

    route: str = "getUnityAirEnvrnIdexSnstiveAboveMsrstnList"


################################################################################
# [API] 시도별 실시간 측정정보 조회 상세기능명세
#
# 버전(ver) 항목설명
# - 버전을 포함하지 않고 호출할 경우 : PM2.5 데이터가 포함되지 않은 원래 오퍼레이션 결과 표출.
# - 버전 1.0을 호출할 경우 : PM2.5 데이터가 포함된 결과 표출.
# - 버전 1.1을 호출할 경우 : PM10, PM2.5 24시간 예측이동 평균데이터가 포함된 결과 표출.
# - 버전 1.2을 호출할 경우 : 측정망 정보 데이터가 포함된 결과 표출.
# - 버전 1.3을 호출할 경우 : PM10, PM2.5 1시간 등급 자료가 포함된 결과 표출
#
# Grade 값의 의미
#   1: 좋음, 2: 보통, 3: 나쁨, 4: 매우나쁨
################################################################################
# Output Model
class CtprvnRltmMesureDnstyModel(BaseModel):
    dataTime: Optional[str]
    stationName: str
    mangName: str
    sidoName: str
    so2Value: Optional[float]
    coValue: Optional[float]
    o3Value: Optional[float]
    no2Value: Optional[float]
    pm10Value: Optional[int]
    pm10Value24: Optional[int]
    pm25Value: Optional[int]
    pm25Value24: Optional[int]
    khaiValue: Optional[int]
    khaiGrade: Optional[int]
    so2Grade: Optional[int]
    coGrade: Optional[int]
    o3Grade: Optional[int]
    no2Grade: Optional[int]
    pm10Grade: Optional[int]
    pm25Grade: Optional[int]
    pm10Grade1h: Optional[int]
    pm25Grade1h: Optional[int]
    so2Flag: Optional[str]
    coFlag: Optional[str]
    o3Flag: Optional[str]
    no2Flag: Optional[str]
    pm10Flag: Optional[str]
    pm25Flag: Optional[str]

    @validator("*", pre=True)
    def _hyphen_to_none(cls, v):
        if v in ["-"]:
            return None
        return v


# API
class CtprvnRltmMesureDnsty(ArpltnInforInqireSvc):
    __RecordModel__ = CtprvnRltmMesureDnstyModel
    __index_names__ = None
    __key_name__ = None
    __value_name__ = None

    route: str = "getCtprvnRltmMesureDnsty"
    sidoName: str
    ver: str = "1.3"


################################################################################
# [API] 대기질 예보통보 조회
#
# 미세먼지/오존 예보 관련 안내사항
#   - 미세먼지 예보는 오늘예보 / 내일예보 / 모레예보가 제공되며, 시시각각으로 변하는 대기질 상황을 전달하기 위해 매일 4회(오전5시, 오전 11시, 오후5시(17시), 오후11시(23시))에 19개 권역으로 발표되고 있습니다.
#   - 내일예보의 경우, 지역별 상세예보는 오전5시, 오전 11시에는 발표되지 않습니다.
#   - 모레예보의 경우, 예보개황만 제공하며 예보등급은 제공하지 않습니다.
#   - 모레예보는 예보 정확도가 낮을 수 있으나 정보 제공을 위해 발표됩니다. (시범운영 : ‘15년 수도권,  ’16년 전국)
#   - 오존예보는 매년 4월15일 ~ 10월15일까지 발표됩니다.
#   - [오늘/내일/모레] 예보 데이터 확인방법
#     "대기질 예보통보 조회" 오퍼레이션 호출 시, 응답 메세지 중 <dataTime>이 2015-12-29 05시 발표일 경우
#     § 오늘예보 : <dataTime> 2015-12-29 05시 발표 <informData> 2015-12-29
#     § 내일예보 : <dataTime> 2015-12-29 05시 발표 <informData> 2015-12-30
#     § 모레예보 : <dataTime> 2015-12-29 05시 발표 <informData> 2015-12-31
################################################################################
# Output Model
class MinuDustFrcstDspthModel(BaseModel):
    dataTime: str
    informCode: str
    informOverall: Optional[str]
    informCause: Optional[str]
    informData: Optional[str]
    informGrade: Optional[str]
    actionKnack: Optional[str]
    imageUrl1: Optional[HttpUrl]
    imageUrl2: Optional[HttpUrl]
    imageUrl3: Optional[HttpUrl]
    imageUrl4: Optional[HttpUrl]
    imageUrl5: Optional[HttpUrl]
    imageUrl6: Optional[HttpUrl]
    imageUrl7: Optional[HttpUrl]
    imageUrl8: Optional[HttpUrl]
    imageUrl9: Optional[HttpUrl]


# API
class MinuDustFrcstDspth(ArpltnInforInqireSvc):
    __RecordModel__ = MinuDustFrcstDspthModel
    __index_names__ = None
    __key_name__ = None
    __value_name__ = None

    route: str = "getMinuDustFrcstDspth"
    searchDate: str = None
    informCode: str = None



import logging
import os
from datetime import datetime
from enum import Enum
from typing import Optional

from pydantic import BaseModel, Field, HttpUrl, SecretStr, ValidationError

from ..DataGoKr import DataGoKr

# logging
logger = logging.getLogger(__file__)

# debug only
KMA_API_KEY = os.getenv("KMA_API_KEY")

################################################################################
# Types
################################################################################
# (Type)
class DataType(str, Enum):
    # Only JSON Available yet
    json = "JSON"


# (Type)
class VilageFcstVersionFtype(str, Enum):
    ODAM = "ODAM"
    VSRT = "VSRT"
    SHRT = "SHRT"


################################################################################
# [Abstract] Abstract for VilageFcst
################################################################################
class VilageFcstInfo(DataGoKr):
    __version__ = "2.0"

    baseUrl: HttpUrl = "http://apis.data.go.kr/1360000/VilageFcstInfoService_2.0"
    dataType: Optional[DataType] = "JSON"  # Only JSON available yet.
    serviceKey: str = KMA_API_KEY


################################################################################
# [API] 초단기 실황  UltraSrtNcst
################################################################################
# Output Model
class UltraSrtNcstModel(BaseModel):
    baseDate: str
    baseTime: str
    T1H: Optional[float]  # 10 decimal
    RN1: Optional[str]  # 8 code
    UUU: Optional[float]  # 12 float
    VVV: Optional[float]  # 12 float
    REH: Optional[int]  # 8 int
    PTY: Optional[int]  # 4 code
    VEC: Optional[float]  # 10 decimal
    WSD: Optional[float]  # 10 decimal


# API
class UltraSrtNcst(VilageFcstInfo):
    __RecordModel__ = UltraSrtNcstModel
    __index_names__ = None
    __key_name__ = "category"
    __value_name__ = "obsrValue"

    route: str = "getUltraSrtNcst"
    base_date: str = datetime.now().strftime("%Y%m%d")
    base_time: str = "0500"
    nx: int = 64
    ny: int = 118


################################################################################
# [API] 초단기 예보 UltraSrtFcst
################################################################################
# Output Model
class UltraSrtFcstModel(BaseModel):
    baseDate: str
    baseTime: str
    fcstDate: str
    fcstTime: str
    T1H: Optional[float]  # 10 decimal
    RN1: Optional[str]  # 8 code
    SKY: Optional[int]  # 4 code
    UUU: Optional[float]  # 12 float
    VVV: Optional[float]  # 12 float
    REH: Optional[int]  # 8 int
    PTY: Optional[int]  # 4 code
    LGT: Optional[str]  # 4 code
    VEC: Optional[float]  # 10 decimal
    WSD: Optional[float]  # 10 decimal


# API
class UltraSrtFcst(VilageFcstInfo):
    __RecordModel__ = UltraSrtFcstModel
    __index_names__ = ["fcstDate", "fcstTime"]
    __key_name__ = "category"
    __value_name__ = "fcstValue"

    route: str = "getUltraSrtFcst"
    base_date: str = datetime.now().strftime("%Y%m%d")
    base_time: str = "0500"
    nx: int = 64
    ny: int = 118


################################################################################
# [API] 단기 예보 VilageFcst
################################################################################
# Output Model
class VilageFcstModel(BaseModel):
    baseDate: str
    baseTime: str
    fcstDate: str
    fcstTime: str
    POP: Optional[int]  # 8 int
    PTY: Optional[int]  # 4 code
    PCP: Optional[str]  # 8 code
    REH: Optional[int]  # 8 int
    SNO: Optional[str]  # 8 code
    SKY: Optional[int]  # 4 code
    TMP: Optional[float]  # 10 decimal
    TMN: Optional[float]  # 10 decimal
    TMX: Optional[float]  # 10 decimal
    UUU: Optional[float]  # 12 float
    VVV: Optional[float]  # 12 float
    WAV: Optional[float]  # 8 int
    VEC: Optional[float]  # 10 decimal
    WSD: Optional[float]  # 10 decimal


# API
class VilageFcst(VilageFcstInfo):
    __RecordModel__ = VilageFcstModel
    __index_names__ = ["fcstDate", "fcstTime"]
    __key_name__ = "category"
    __value_name__ = "fcstValue"

    route: str = "getVilageFcst"
    base_date: str = datetime.now().strftime("%Y%m%d")
    base_time: str = "0500"
    nx: int = 64
    ny: int = 118


################################################################################
# [API] 단기예보 수치모델 버전
################################################################################
# Output Model
class VilageFcstVersion(BaseModel):
    filetype: VilageFcstVersionFtype
    version: str


# API
class VilageFcstVersion(VilageFcstInfo):
    __RecordModel__ = VilageFcstVersion
    route: str = "getFcstVersion"
    ftype: VilageFcstVersionFtype = "ODAM"
    basedatetime: str = datetime.now().strftime("%Y%m%d0800")



import json

import click

from .airkorea import MsrstnAcctoRltmMesureDnsty
from .kma import UltraSrtFcst, UltraSrtNcst, VilageFcst, WthrDataList


def _out(records):
    for record in records:
        print(record.dict())


@click.group()
def datagokr():
    pass


@datagokr.command()
@click.option("-s", "--service-key", default=None, help="API Key")
@click.option("-d", "--base-date", default=None, help="baseDate")
@click.option("-t", "--base-time", default=None, help="baseTime")
@click.option("-x", "--nx", default=None, help="nx")
@click.option("-y", "--ny", default=None, help="ny")
def ultra_srt_ncst(service_key, base_date, base_time, nx, ny):
    params = {
        "serviceKey": service_key,
        "base_date": base_date,
        "base_time": base_time,
        "nx": nx,
        "ny": ny,
    }
    api = UltraSrtNcst(**{k: v for k, v in params.items() if v is not None})
    records = api.get_records()
    _out(records)


@datagokr.command()
@click.option("-s", "--service-key", default=None, help="API Key")
@click.option("-d", "--base-date", default=None, help="baseDate")
@click.option("-t", "--base-time", default=None, help="baseTime")
@click.option("-x", "--nx", default=None, help="nx")
@click.option("-y", "--ny", default=None, help="ny")
def ultra_srt_fcst(service_key, base_date, base_time, nx, ny):
    params = {
        "serviceKey": service_key,
        "base_date": base_date,
        "base_time": base_time,
        "nx": nx,
        "ny": ny,
    }
    api = UltraSrtFcst(**{k: v for k, v in params.items() if v is not None})
    records = api.get_records()
    _out(records)


@datagokr.command()
@click.option("-s", "--service-key", default=None, help="API Key")
@click.option("-d", "--base-date", default=None, help="baseDate")
@click.option("-t", "--base-time", default=None, help="baseTime")
@click.option("-x", "--nx", default=None, help="nx")
@click.option("-y", "--ny", default=None, help="ny")
def vilage_fcst(service_key, base_date, base_time, nx, ny):
    params = {
        "serviceKey": service_key,
        "base_date": base_date,
        "base_time": base_time,
        "nx": nx,
        "ny": ny,
    }
    api = VilageFcst(**{k: v for k, v in params.items() if v is not None})
    records = api.get_records()
    _out(records)



import logging
import os
from datetime import datetime, timedelta
from enum import Enum
from typing import Optional

from pydantic import BaseModel, Field, HttpUrl, SecretStr, ValidationError

from ..DataGoKr import DataGoKr

# logging
logger = logging.getLogger(__file__)

# debug only
KMA_API_KEY = os.getenv("KMA_API_KEY")


################################################################################
# Types
################################################################################
class DataType(str, Enum):
    # Only JSON Available yet
    json = "JSON"


################################################################################
# [Abstract API] 지상(종관,ASOS)
################################################################################
# ABSTRACT
class AsosHourlyInfo(DataGoKr):
    __version__ = "2.0"

    baseUrl: HttpUrl = "http://apis.data.go.kr/1360000/AsosHourlyInfoService"
    dataType: Optional[DataType] = "JSON"  # Only JSON available yet.
    serviceKey: str = KMA_API_KEY


################################################################################
# [API] 지상(종관,ASOS) 시간자료 조회
################################################################################
# Output Model
class WthrDataListModel(BaseModel):
    tm: str  # 시간
    rnum: int  # 목록순서
    stnId: int  # 지점번호
    stnNm: str  # 지점이름
    ta: Optional[float]  # 기온
    taQcflg: Optional[int]  # 기온품질검사플래그 (null: 정상, 1: 오류, 9: 결측)
    rn: Optional[float]  # 강수량
    rnQcflg: Optional[int]
    ws: Optional[float]  # 풍속
    wsQcflg: Optional[int]
    wd: Optional[str]  # 풍향 (20: 북북동, 40: 북동, ..., 360: 정북, 00: CALM, 99: 변화많음)
    wdQcflg: Optional[int]
    hm: Optional[int]  # 상대습도
    hmQcflg: Optional[int]
    pv: Optional[float]  # 증기압
    td: Optional[float]  # 이슬점온도
    pa: Optional[float]  # 현지기압
    paQcflg: Optional[int]
    ps: Optional[float]  # 해면기압
    psQcflg: Optional[int]
    ss: Optional[float]  # 일조시간
    ssQcflg: Optional[int]
    icsr: Optional[float]  # 일사량
    dsnw: Optional[float]  # 적설량
    hr3Fhsc: Optional[float]  # 3시간신적설
    dc10Tca: Optional[int]  # 전운량
    dc10LmcsCa: Optional[int]  # 중하층운량
    clfmAbbrCd: Optional[str]  # 운형
    lcsCh: Optional[int]  # 최저운고
    vs: Optional[int]  # 시정
    # gndSttCd: Optional[int]   # 2016.7.1 종료
    dmstMtphNo: Optional[str]  # 현상번호
    ts: Optional[float]  # 지면온도
    tsQcflg: Optional[int]
    m005Te: Optional[float]  # 5cm 지중온도
    m01Te: Optional[float]  # 10cm 지중온도
    m02Te: Optional[float]  # 20cm 지중온도
    m03Te: Optional[float]  # 30cm 지중온도


# API
class WthrDataList(AsosHourlyInfo):
    __RecordModel__ = WthrDataListModel
    __index_names__ = None
    __key_name__ = None
    __value_name__ = None

    route: str = "getWthrDataList"
    dataCd: str = "ASOS"
    dateCd: str = "HR"
    startDt: str = (datetime.now() - timedelta(days=2)).strftime("%Y%m%d")
    startHh: str = (datetime.now() - timedelta(days=2)).strftime("%H")
    endDt: str = (datetime.now() - timedelta(days=1)).strftime("%Y%m%d")
    endHh: str = (datetime.now() - timedelta(days=1)).strftime("%H")
    stnIds: int = 119  # 이동읍 최근접 - 수원 수도권기상청



import abc
import logging
import os
from typing import List, Optional
from urllib.parse import quote_plus, unquote, urlencode

import pendulum
import requests
from pydantic import BaseModel, Field, HttpUrl, SecretStr, ValidationError

# logging
logger = logging.getLogger(__file__)

# timezone
KST = pendulum.timezone("Asia/Seoul")


################################################################################
# Mother Class
################################################################################
class DataGoKr(BaseModel, abc.ABC):
    """
    (NOTE)
     - numOfRows: 데이터 요청은 1,000건을 넘을 수 없음. (MAX 999)
     - pageNo: 디버그 시에만 사용.
    """

    __timezone__: KST
    __RecordModel__: BaseModel = None
    __index_names__: list = None
    __key_name__: str = None
    __value_name__: str = None

    baseUrl: HttpUrl
    route: str
    serviceKey: SecretStr
    numOfRows: Optional[int] = 999
    pageNo: Optional[int]

    class Config:
        use_enum_values = True

    def request(self, pageNo=None):
        if pageNo is not None:
            self.pageNo = pageNo

        # request and receive response
        endpoint = self.get_endpoint()
        response = self._request(endpoint=endpoint)
        if response["header"]["resultCode"] != "00":
            _msg = f"CODE[{response['header']['resultCode']}]: {response['header']['resultMsg']}"
            raise ReferenceError(_msg)

        # extract records from body
        body = response["body"]
        records = body["items"]
        if not isinstance(records, list):
            records = records["item"]

        # repeat request, extract, and append records until the last page
        if pageNo is None:
            self.pageNo = 1
            while body["totalCount"] >= body["numOfRows"]:
                self.pageNo = body["pageNo"] + 1
                endpoint = self.get_endpoint()
                response = self._request(endpoint=endpoint)
                # result code '03' means 'No data'
                if response["header"]["resultCode"] == "03":
                    break
                if response["header"]["resultCode"] != "00":
                    _msg = f"CODE[{response['header']['resultCode']}]: {response['header']['resultMsg']}"
                    raise ReferenceError(_msg)
                body = response["body"]
                records += body["items"]["item"]
        return records

    def get_endpoint(self):
        url = os.path.join(self.baseUrl, self.route)
        query = urlencode(
            {quote_plus(k): v for k, v in self.__dict__.items() if k not in ["baseUrl", "route"] and v is not None}
        )
        return "?".join([url, query])

    def check_query(self):
        return {k: v for k, v in self.__dict__.items() if not k.startswith("_")}

    def get_records(self, sep="___"):
        items = self.request()
        items = list(map(self._drop_empty_fields, items))

        # drop invalid items
        items = [_ for _ in items if _ is not None]

        if self.__key_name__ is None or self.__value_name__ is None:
            return [self.__RecordModel__(**item) for item in items]

        records = self._items_to_records(
            Model=self.__RecordModel__,
            items=items,
            index_names=self.__index_names__,
            key_name=self.__key_name__,
            value_name=self.__value_name__,
            sep=sep,
        )

        if self.__RecordModel__ is None:
            return records

        records = [self._record_to_model(self.__RecordModel__, record) for record in records]
        return [r for r in records if r is not None]

    @staticmethod
    def _request(endpoint):
        response = requests.get(endpoint)
        if response.text.startswith("<"):
            _msg = f"[COMMON ERROR] XML is returned - {response.text}"
            raise ReferenceError(_msg)
        response = response.json()["response"]
        return response

    @staticmethod
    def _items_to_records(Model, items, key_name, value_name, index_names, sep):

        # required fields
        required = Model.schema().get("required", [])

        if index_names is None:
            record = dict()
            for item in items:
                record.update({item.pop(key_name): item.pop(value_name)})
                record.update(item)
            return [record]

        records = dict()
        for item in items:
            flat_indices = sep.join([item.pop(idx) for idx in index_names])
            if flat_indices not in records.keys():
                records[flat_indices] = dict()
            records[flat_indices].update({item.pop(key_name): item.pop(value_name)})
            records[flat_indices].update(item)

        return [
            {**{k: v for k, v in zip(index_names, flat_indices.split(sep))}, **record}
            for flat_indices, record in records.items()
        ]

    @staticmethod
    def _record_to_model(Model, record):
        try:
            return Model(**record)
        except ValidationError as ve:
            _msg = "_".join([f"{k}:{v}" for k, v in record.items()])
            logger.warn(f"{ve} - {_msg}")

    @staticmethod
    def _drop_empty_fields(item: dict):
        # drop empty fields
        return {k: v for k, v in item.items() if v != ""}

    @staticmethod
    def _ensure_list(x):
        if x is None or isinstance(x, list):
            return x
        return [x]
